import numpy as np
import pytest
from numpy.testing import assert_array_equal
import scipy.ndimage as ndi

from skimage import io, draw
from skimage._shared.testing import fetch
from skimage.data import binary_blobs
from skimage.morphology import medial_axis, skeletonize, thin
from skimage.morphology._skeletonize import G123_LUT, G123P_LUT, _generate_thin_luts


class TestSkeletonize:
    @pytest.mark.parametrize("method", ["zhang", "lee"])
    def test_no_foreground(self, method):
        image = np.zeros((5, 5))
        result = skeletonize(image, method=method)
        assert_array_equal(result, np.zeros((5, 5)))

    @pytest.mark.parametrize(
        "ndim,method", [(1, "zhang"), (3, "zhang"), (1, "lee"), (4, "lee")]
    )
    def test_wrong_ndim(self, ndim, method):
        image = np.zeros((5,) * ndim, dtype=bool)
        with pytest.raises(ValueError):
            skeletonize(image, method=method)

    def test_wrong_method(self):
        image = np.ones((5, 5), dtype=bool)
        with pytest.raises(ValueError):
            skeletonize(image, method="foo")

    @pytest.mark.parametrize("method", ["zhang", "lee"])
    def test_skeletonize_all_foreground(self, method):
        image = np.ones((3, 4), dtype=bool)
        result = skeletonize(image, method=method)
        if method == "zhang":
            expected = np.array([[0, 0, 1, 0], [1, 1, 0, 0], [0, 0, 0, 0]], dtype=bool)
        else:  # "lee"
            expected = np.array([[0, 0, 0, 0], [1, 1, 1, 1], [0, 0, 0, 0]], dtype=bool)
        assert_array_equal(result, expected)

    @pytest.mark.parametrize("method", ["zhang", "lee"])
    def test_single_point(self, method):
        image = np.zeros((5, 5), dtype=bool)
        image[3, 3] = 1
        result = skeletonize(image, method=method)
        assert_array_equal(result, image)

    @pytest.mark.parametrize("method", ["zhang", "lee"])
    def test_vec_1d(self, method):
        # Corner case of a 2D image, which is a 1D vector
        image = np.ones((5, 1), dtype=bool)
        result = skeletonize(image, method=method)
        assert_array_equal(result, image)

    @pytest.mark.parametrize("method", ["zhang", "lee"])
    def test_already_thinned(self, method):
        image = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1],
                [0, 1, 1, 1, 0],
                [1, 0, 0, 0, 0],
            ],
            dtype=bool,
        )
        result = skeletonize(image, method=method)
        assert_array_equal(result, image)

    def test_output(self):
        image = io.imread(fetch("data/bw_text.png"), as_gray=True)

        # make black the foreground
        image = image == 0
        result = skeletonize(image)

        expected = np.load(fetch("data/bw_text_skeleton.npy"))
        assert_array_equal(result, expected)

    @pytest.mark.parametrize("method", ["zhang", "lee"])
    @pytest.mark.parametrize("dtype", [bool, float, int])
    def test_num_neighbors(self, method, dtype):
        # an empty image
        image = np.zeros((300, 300), dtype=dtype)

        # foreground object 1
        image[10:-10, 10:100] = 1
        image[-100:-10, 10:-10] = 2
        image[10:-10, -100:-10] = 3

        # foreground object 2
        rs, cs = draw.line(250, 150, 10, 280)
        for i in range(10):
            image[rs + i, cs] = 4
        rs, cs = draw.line(10, 150, 250, 280)
        for i in range(20):
            image[rs + i, cs] = 5

        # foreground object 3
        ir, ic = np.indices(image.shape)
        circle1 = (ic - 135) ** 2 + (ir - 150) ** 2 < 30**2
        circle2 = (ic - 135) ** 2 + (ir - 150) ** 2 < 20**2
        image[circle1] = 1
        image[circle2] = 0
        result = skeletonize(image, method=method).astype(np.uint8)

        # there should never be a 2x2 block of foreground pixels in a skeleton
        mask = np.array([[1, 1], [1, 1]], np.uint8)
        blocks = ndi.correlate(result, mask, mode="constant")
        assert not np.any(blocks == 4)

    def test_lut_fix(self):
        image = np.array(
            [
                [0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0],
                [0, 0, 1, 1, 0, 0],
                [0, 0, 0, 1, 1, 0],
                [0, 0, 0, 0, 1, 1],
                [0, 0, 0, 0, 0, 0],
            ],
            dtype=bool,
        )
        result = skeletonize(image)
        expected = np.array(
            [
                [0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0],
                [0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0],
            ],
            dtype=bool,
        )
        assert np.all(result == expected)

    @pytest.mark.parametrize("ndim,method", [(2, "zhang"), (2, "lee"), (3, "lee")])
    @pytest.mark.parametrize("dtype", [bool, np.uint8])
    def test_input_not_modified(self, method, ndim, dtype):
        # Skeletonize must not modify the input image
        image = np.ones((3,) * ndim, dtype=dtype)
        image = np.pad(image, 1)
        original = image.copy()
        _ = skeletonize(image, method=method)
        np.testing.assert_array_equal(image, original)

    @pytest.mark.parametrize("method", ["zhang", "lee"])
    def test_input_float_conv(self, method):
        # Check that the floats are correctly handled. Also check non-contiguous input
        image = np.random.random((16, 16))[::2, ::2]
        image[image < 0.5] = 0.0

        original = image.copy()
        result = skeletonize(image, method=method)

        assert result.dtype == bool
        assert_array_equal(image, original)

    def test_two_hole_image_vs_fiji(self):
        # Test a simple 2D image against FIJI
        image = np.array(
            [
                [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0],
                [0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0],
                [0, 0, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0],
                [0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0],
                [0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ],
            dtype=bool,
        )
        expected = np.array(
            [
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ],
            dtype=bool,
        )
        result = skeletonize(image, method="lee")
        assert_array_equal(result, expected)

    def test_3d_vs_fiji(self):
        # Generate an image with blobs and compare its skeleton
        # to the one generated by FIJI (Plugins>Skeleton->Skeletonize)
        image = binary_blobs(32, 0.05, n_dim=3, rng=1234)
        image = image[:-2, ...]

        result = skeletonize(image)
        expected = io.imread(fetch("data/_blobs_3d_fiji_skeleton.tif")).astype(bool)
        assert_array_equal(result, expected)


class TestThin:
    @property
    def input_image(self):
        # Image to test thinning with
        ii = np.array(
            [
                [0, 0, 0, 0, 0, 0, 0],
                [0, 1, 2, 3, 4, 5, 0],
                [0, 1, 0, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 6, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 0, 0, 0, 0, 0, 0],
            ],
            dtype=float,
        )
        return ii

    def test_all_zeros(self):
        image = np.zeros((10, 10), dtype=bool)
        assert np.all(thin(image) == False)

    @pytest.mark.parametrize("dtype", [bool, float, int])
    def test_thin_copies_input(self, dtype):
        """Ensure thinning does not modify the input image."""
        image = self.input_image.astype(dtype)
        original = image.copy()
        thin(image)
        np.testing.assert_array_equal(image, original)

    @pytest.mark.parametrize("dtype", [bool, float, int])
    def test_iter_1(self, dtype):
        image = self.input_image.astype(dtype)
        result = thin(image, 1).astype(bool)
        expected = np.array(
            [
                [0, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0],
                [0, 1, 0, 1, 1, 0, 0],
                [0, 0, 1, 1, 1, 0, 0],
                [0, 0, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
            ],
            dtype=bool,
        )
        assert_array_equal(result, expected)

    @pytest.mark.parametrize("dtype", [bool, float, int])
    def test_noiter(self, dtype):
        image = self.input_image.astype(dtype)
        result = thin(image).astype(bool)
        expected = np.array(
            [
                [0, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0],
                [0, 1, 0, 1, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
            ],
            dtype=bool,
        )
        assert_array_equal(result, expected)

    def test_baddim(self):
        for ii in [np.zeros(3, dtype=bool), np.zeros((3, 3, 3), dtype=bool)]:
            with pytest.raises(ValueError):
                thin(ii)

    def test_lut_generation(self):
        g123, g123p = _generate_thin_luts()

        assert_array_equal(g123, G123_LUT)
        assert_array_equal(g123p, G123P_LUT)


class TestMedialAxis:
    def test_all_zeros(self):
        result = medial_axis(np.zeros((10, 10), dtype=bool))
        assert np.all(result == False)

    def test_all_zeros_masked(self):
        result = medial_axis(
            np.zeros((10, 10), dtype=bool), np.zeros((10, 10), dtype=bool)
        )
        assert np.all(result == False)

    @pytest.mark.parametrize("dtype", [bool, float, int])
    def test_vertical_line(self, dtype):
        # Image is a thick vertical line (see gh-3861)
        image = np.zeros((9, 9), dtype=dtype)
        image[:, 2] = 1
        image[:, 3] = 2
        image[:, 4] = 3

        expected = np.full(image.shape, False)
        expected[:, 3] = True

        result = medial_axis(image)
        assert_array_equal(result, expected)

    def test_rectangle(self):
        image = np.zeros((9, 15), dtype=bool)
        image[1:-1, 1:-1] = True
        # Excepted are four diagonals from the corners, meeting in a horizontal line
        expected = np.array(
            [
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ],
            dtype=bool,
        )
        result = medial_axis(image)
        assert np.all(result == expected)
        result, distance = medial_axis(image, return_distance=True)
        assert distance.max() == 4

    def test_rectange_with_hole(self):
        image = np.zeros((9, 15), dtype=bool)
        image[1:-1, 1:-1] = True
        image[4, 4:-4] = False
        expected = np.array(
            [
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ],
            dtype=bool,
        )
        result = medial_axis(image)
        assert np.all(result == expected)

    def test_narrow_image(self):
        # Image is a 1-pixel thin strip
        image = np.zeros((1, 5), dtype=bool)
        image[:, 1:-1] = True
        result = medial_axis(image)
        assert np.all(result == image)



import numpy as np

from .._shared.utils import _supported_float_type
from ._rolling_ball_cy import apply_kernel, apply_kernel_nan


def rolling_ball(image, *, radius=100, kernel=None, nansafe=False, num_threads=None):
    """Estimate background intensity using the rolling-ball algorithm.

    This function is a generalization of the rolling-ball algorithm [1]_ to
    estimate the background intensity of an n-dimensional image. This is
    typically useful for background subtraction in case of uneven exposure.
    Think of the image as a landscape (where altitude is determined by
    intensity), under which a ball of given radius is rolled. At each
    position, the ball's apex gives the resulting background intensity.

    Parameters
    ----------
    image : ndarray
        The image to be filtered.
    radius : int, optional
        Radius of the ball-shaped kernel to be rolled under the
        image landscape. Used only if `kernel` is ``None``.
    kernel : ndarray, optional
        An alternative way to specify the rolling ball, as an arbitrary
        kernel. It must have the same number of axes as `image`.
    nansafe: bool, optional
        If ``False`` (default), the function assumes that none of the values
        in `image` are ``np.nan``, and uses a faster implementation.
    num_threads: int, optional
        The maximum number of threads to use. If ``None``, the function uses
        the OpenMP default value; typically, it is equal to the maximum number
        of virtual cores.
        Note: This is an upper limit to the number of threads. The exact number
        is determined by the system's OpenMP library.

    Returns
    -------
    background : ndarray
        The estimated background of the image.

    Notes
    -----
    This implementation assumes that dark pixels correspond to the background. If
    you have a bright background, invert the image before passing it to this
    function, e.g., using :func:`skimage.util.invert`.

    For this method to give meaningful results, the radius of the ball (or
    typical size of the kernel, in the general case) should be larger than the
    typical size of the image features of interest.

    This algorithm is sensitive to noise (in particular salt-and-pepper
    noise). If this is a problem in your image, you can apply mild
    Gaussian smoothing before passing the image to this function.

    This algorithm's complexity is polynomial in the radius, with degree equal
    to the image dimensionality (a 2D image is N^2, a 3D image is N^3, etc.),
    so it can take a long time as the radius grows beyond 30 or so ([2]_, [3]_).
    It is an exact N-dimensional calculation; if all you need is an
    approximation, faster options to consider are top-hat filtering [4]_ or
    downscaling-then-upscaling to reduce the size of the input processed.

    References
    ----------
    .. [1] Sternberg, Stanley R. "Biomedical image processing." Computer 1
           (1983): 22-34. :DOI:`10.1109/MC.1983.1654163`
    .. [2] https://github.com/scikit-image/scikit-image/issues/5193
    .. [3] https://github.com/scikit-image/scikit-image/issues/7423
    .. [4] https://forum.image.sc/t/59267/7

    Examples
    --------
    >>> import numpy as np
    >>> import skimage as ski
    >>> image = ski.data.coins()
    >>> background = ski.restoration.rolling_ball(image)
    >>> filtered_image = image - background

    >>> import numpy as np
    >>> import skimage as ski
    >>> image = ski.data.coins()
    >>> kernel = ski.restoration.ellipsoid_kernel((101, 101), 75)
    >>> background = ski.restoration.rolling_ball(image, kernel=kernel)
    >>> filtered_image = image - background
    """

    image = np.asarray(image)
    float_type = _supported_float_type(image.dtype)
    img = image.astype(float_type, copy=False)

    if num_threads is None:
        num_threads = 0

    if kernel is None:
        kernel = ball_kernel(radius, image.ndim)

    kernel = kernel.astype(float_type)
    kernel_shape = np.asarray(kernel.shape)
    kernel_center = kernel_shape // 2
    center_intensity = kernel[tuple(kernel_center)]

    intensity_difference = center_intensity - kernel
    intensity_difference[kernel == np.inf] = np.inf
    intensity_difference = intensity_difference.astype(img.dtype)
    intensity_difference = intensity_difference.reshape(-1)

    img = np.pad(
        img, kernel_center[:, np.newaxis], constant_values=np.inf, mode="constant"
    )

    func = apply_kernel_nan if nansafe else apply_kernel
    background = func(
        img.reshape(-1),
        intensity_difference,
        np.zeros_like(image, dtype=img.dtype).reshape(-1),
        np.array(image.shape, dtype=np.intp),
        np.array(img.shape, dtype=np.intp),
        kernel_shape.astype(np.intp),
        num_threads,
    )

    background = background.astype(image.dtype, copy=False)

    return background


def ball_kernel(radius, ndim):
    """Create a ball kernel for restoration.rolling_ball.

    Parameters
    ----------
    radius : int
        Radius of the ball.
    ndim : int
        Number of dimensions of the ball. ``ndim`` should match the
        dimensionality of the image the kernel will be applied to.

    Returns
    -------
    kernel : ndarray
        The kernel containing the surface intensity of the top half
        of the ellipsoid.

    See Also
    --------
    rolling_ball
    """

    kernel_coords = np.stack(
        np.meshgrid(
            *[np.arange(-x, x + 1) for x in [np.ceil(radius)] * ndim], indexing='ij'
        ),
        axis=-1,
    )

    sum_of_squares = np.sum(kernel_coords**2, axis=-1)
    distance_from_center = np.sqrt(sum_of_squares)
    kernel = np.sqrt(np.clip(radius**2 - sum_of_squares, 0, None))
    kernel[distance_from_center > radius] = np.inf

    return kernel


def ellipsoid_kernel(shape, intensity):
    """Create an ellipoid kernel for restoration.rolling_ball.

    Parameters
    ----------
    shape : array-like
        Length of the principal axis of the ellipsoid (excluding
        the intensity axis). The kernel needs to have the same
        dimensionality as the image it will be applied to.
    intensity : int
        Length of the intensity axis of the ellipsoid.

    Returns
    -------
    kernel : ndarray
        The kernel containing the surface intensity of the top half
        of the ellipsoid.

    See Also
    --------
    rolling_ball
    """

    shape = np.asarray(shape)
    semi_axis = np.clip(shape // 2, 1, None)

    kernel_coords = np.stack(
        np.meshgrid(*[np.arange(-x, x + 1) for x in semi_axis], indexing='ij'), axis=-1
    )

    intensity_scaling = 1 - np.sum((kernel_coords / semi_axis) ** 2, axis=-1)
    kernel = intensity * np.sqrt(np.clip(intensity_scaling, 0, None))
    kernel[intensity_scaling < 0] = np.inf

    return kernel



"""Morphological algorithms, e.g., closing, opening, skeletonization."""

from .binary import binary_closing, binary_dilation, binary_erosion, binary_opening
from .gray import black_tophat, closing, dilation, erosion, opening, white_tophat
from .isotropic import (
    isotropic_erosion,
    isotropic_dilation,
    isotropic_opening,
    isotropic_closing,
)
from .footprints import (
    ball,
    cube,
    diamond,
    disk,
    ellipse,
    footprint_from_sequence,
    footprint_rectangle,
    mirror_footprint,
    octagon,
    octahedron,
    pad_footprint,
    rectangle,
    square,
    star,
)
from ..measure._label import label
from ._skeletonize import medial_axis, skeletonize, thin
from .convex_hull import convex_hull_image, convex_hull_object
from .grayreconstruct import reconstruction
from .misc import remove_small_holes, remove_small_objects, remove_objects_by_distance
from .extrema import h_maxima, h_minima, local_minima, local_maxima
from ._flood_fill import flood, flood_fill
from .max_tree import (
    area_opening,
    area_closing,
    diameter_closing,
    diameter_opening,
    max_tree,
    max_tree_local_maxima,
)

__all__ = [
    'area_closing',
    'area_opening',
    'ball',
    'binary_closing',
    'binary_dilation',
    'binary_erosion',
    'binary_opening',
    'black_tophat',
    'closing',
    'convex_hull_image',
    'convex_hull_object',
    'diameter_closing',
    'diameter_opening',
    'diamond',
    'dilation',
    'disk',
    'ellipse',
    'erosion',
    'flood',
    'flood_fill',
    'footprint_from_sequence',
    'footprint_rectangle',
    'h_maxima',
    'h_minima',
    'isotropic_closing',
    'isotropic_dilation',
    'isotropic_erosion',
    'isotropic_opening',
    'label',
    'local_maxima',
    'local_minima',
    'max_tree',
    'max_tree_local_maxima',
    'medial_axis',
    'mirror_footprint',
    'octagon',
    'octahedron',
    'opening',
    'pad_footprint',
    'reconstruction',
    'remove_small_holes',
    'remove_small_objects',
    'remove_objects_by_distance',
    'skeletonize',
    'star',
    'thin',
    'white_tophat',
]



import skimage as ski


def test_lookfor_basic(capsys):
    assert ski.lookfor is ski.util.lookfor

    ski.util.lookfor("regionprops")
    search_results = capsys.readouterr().out
    assert "skimage.measure.regionprops" in search_results
    assert "skimage.measure.regionprops_table" in search_results



"""Image Processing for Python

scikit-image (a.k.a. ``skimage``) is a collection of algorithms for image
processing and computer vision.

Attributes
----------
__version__ : str
    The scikit-image version string.

Subpackages
-----------
color
    Color space conversion.
data
    Example images and datasets.
draw
    Drawing primitives, such as lines, circles, text, etc.
exposure
    Image intensity adjustment, e.g., histogram equalization, etc.
feature
    Feature detection and extraction, e.g., texture analysis, corners, etc.
filters
    Sharpening, edge finding, rank filters, thresholding, etc.
future
    Functionality with an experimental API.
graph
    Graph-based operations, e.g., shortest paths.
io
    Reading and saving of images and videos.
measure
    Measurement of image properties, e.g., region properties, contours.
metrics
    Metrics corresponding to images, e.g., distance metrics, similarity, etc.
morphology
    Morphological algorithms, e.g., closing, opening, skeletonization.
registration
    Image registration algorithms, e.g., optical flow or phase cross correlation.
restoration
    Restoration algorithms, e.g., deconvolution algorithms, denoising, etc.
segmentation
    Algorithms to partition images into meaningful regions or boundaries.
transform
    Geometric and other transformations, e.g., rotations, Radon transform.
util
    Generic utilities.
"""

__version__ = '0.25.2'

import lazy_loader as _lazy

__getattr__, *_ = _lazy.attach_stub(__name__, __file__)


# Don't use the `__all__` and `__dir__` returned by `attach_stubs` since that
# one would expose utility functions we don't want to advertise in our
# top-level module anymore.
__all__ = [
    "__version__",
    "color",
    "data",
    "draw",
    "exposure",
    "feature",
    "filters",
    "future",
    "graph",
    "io",
    "measure",
    "metrics",
    "morphology",
    "registration",
    "restoration",
    "segmentation",
    "transform",
    "util",
]


def __dir__():
    return __all__.copy()


# Logic for checking for improper install and importing while in the source
# tree when package has not been installed inplace.
# Code adapted from scikit-learn's __check_build module.
_INPLACE_MSG = """
It appears that you are importing a local scikit-image source tree. For
this, you need to have an inplace install. Maybe you are in the source
directory and you need to try from another location."""

_STANDARD_MSG = """
Your install of scikit-image appears to be broken.
Try re-installing the package following the instructions at:
https://scikit-image.org/docs/stable/user_guide/install.html"""


def _raise_build_error(e):
    # Raise a comprehensible error
    import os.path as osp

    local_dir = osp.split(__file__)[0]
    msg = _STANDARD_MSG
    if local_dir == "skimage":
        # Picking up the local install: this will work only if the
        # install is an 'inplace build'
        msg = _INPLACE_MSG
    raise ImportError(
        f"{e}\nIt seems that scikit-image has not been built correctly.\n{msg}"
    )


def _try_append_commit_info(version):
    """Append last commit date and hash to `version`, if available."""
    import subprocess
    from pathlib import Path

    try:
        output = subprocess.check_output(
            ['git', 'log', '-1', '--format="%h %aI"'],
            cwd=Path(__file__).parent,
            text=True,
        )
        if output:
            git_hash, git_date = (
                output.strip().replace('"', '').split('T')[0].replace('-', '').split()
            )
            version = '+'.join(
                [tag for tag in version.split('+') if not tag.startswith('git')]
            )
            version += f'+git{git_date}.{git_hash}'

    except (FileNotFoundError, subprocess.CalledProcessError):
        pass
    except OSError:
        pass  # If skimage is built with emscripten which does not support processes

    return version


if 'dev' in __version__:
    __version__ = _try_append_commit_info(__version__)


from skimage._shared.tester import PytestTester as _PytestTester

test = _PytestTester(__name__)



import sys

from .._vendored.numpy_lookfor import lookfor as _lookfor


def lookfor(what):
    """Do a keyword search on scikit-image docstrings and print results.

    .. warning::

        This function may also print results that are not part of
        scikit-image's public API.

    Parameters
    ----------
    what : str
        Words to look for.

    Examples
    --------
    >>> import skimage as ski
    >>> ski.util.lookfor('regular_grid')
    Search results for 'regular_grid'
    ---------------------------------
    skimage.util.regular_grid
        Find `n_points` regularly spaced along `ar_shape`.
    skimage.util.lookfor
        Do a keyword search on scikit-image docstrings and print results.
    """
    return _lookfor(what, sys.modules[__name__.split('.')[0]])



import pathlib
import warnings

import numpy as np

from .._shared.utils import warn, deprecate_func, deprecate_parameter, DEPRECATED
from .._shared.version_requirements import require
from ..exposure import is_low_contrast
from ..color.colorconv import rgb2gray, rgba2rgb
from ..io.manage_plugins import call_plugin, _hide_plugin_deprecation_warnings
from .util import file_or_url_context

__all__ = [
    'imread',
    'imsave',
    'imshow',
    'show',
    'imread_collection',
    'imshow_collection',
]


_remove_plugin_param_template = (
    "The plugin infrastructure in `skimage.io` and the parameter "
    "`{deprecated_name}` are deprecated since version {deprecated_version} and "
    "will be removed in {changed_version} (or later). To avoid this warning, "
    "please do not use the parameter `{deprecated_name}`. Instead, use `imageio` "
    "or other I/O packages directly. See also `{func_name}`."
)


@deprecate_parameter(
    "plugin",
    start_version="0.25",
    stop_version="0.27",
    template=_remove_plugin_param_template,
)
def imread(fname, as_gray=False, plugin=DEPRECATED, **plugin_args):
    """Load an image from file.

    Parameters
    ----------
    fname : str or pathlib.Path
        Image file name, e.g. ``test.jpg`` or URL.
    as_gray : bool, optional
        If True, convert color images to gray-scale (64-bit floats).
        Images that are already in gray-scale format are not converted.

    Other Parameters
    ----------------
    plugin_args : DEPRECATED
        The plugin infrastructure is deprecated.

    Returns
    -------
    img_array : ndarray
        The different color bands/channels are stored in the
        third dimension, such that a gray-image is MxN, an
        RGB-image MxNx3 and an RGBA-image MxNx4.

    """
    if plugin is DEPRECATED:
        plugin = None
    if plugin_args:
        msg = (
            "The plugin infrastructure in `skimage.io` is deprecated since "
            "version 0.25 and will be removed in 0.27 (or later). To avoid "
            "this warning, please do not pass additional keyword arguments "
            "for plugins (`**plugin_args`). Instead, use `imageio` or other "
            "I/O packages directly. See also `skimage.io.imread`."
        )
        warnings.warn(msg, category=FutureWarning, stacklevel=3)

    if isinstance(fname, pathlib.Path):
        fname = str(fname.resolve())

    if plugin is None and hasattr(fname, 'lower'):
        if fname.lower().endswith(('.tiff', '.tif')):
            plugin = 'tifffile'

    with file_or_url_context(fname) as fname, _hide_plugin_deprecation_warnings():
        img = call_plugin('imread', fname, plugin=plugin, **plugin_args)

    if not hasattr(img, 'ndim'):
        return img

    if img.ndim > 2:
        if img.shape[-1] not in (3, 4) and img.shape[-3] in (3, 4):
            img = np.swapaxes(img, -1, -3)
            img = np.swapaxes(img, -2, -3)

        if as_gray:
            if img.shape[2] == 4:
                img = rgba2rgb(img)
            img = rgb2gray(img)

    return img


@deprecate_parameter(
    "plugin",
    start_version="0.25",
    stop_version="0.27",
    template=_remove_plugin_param_template,
)
def imread_collection(
    load_pattern, conserve_memory=True, plugin=DEPRECATED, **plugin_args
):
    """
    Load a collection of images.

    Parameters
    ----------
    load_pattern : str or list
        List of objects to load. These are usually filenames, but may
        vary depending on the currently active plugin. See :class:`ImageCollection`
        for the default behaviour of this parameter.
    conserve_memory : bool, optional
        If True, never keep more than one in memory at a specific
        time.  Otherwise, images will be cached once they are loaded.

    Returns
    -------
    ic : :class:`ImageCollection`
        Collection of images.

    Other Parameters
    ----------------
    plugin_args : DEPRECATED
        The plugin infrastructure is deprecated.

    """
    if plugin is DEPRECATED:
        plugin = None
    if plugin_args:
        msg = (
            "The plugin infrastructure in `skimage.io` is deprecated since "
            "version 0.25 and will be removed in 0.27 (or later). To avoid "
            "this warning, please do not pass additional keyword arguments "
            "for plugins (`**plugin_args`). Instead, use `imageio` or other "
            "I/O packages directly. See also `skimage.io.imread_collection`."
        )
        warnings.warn(msg, category=FutureWarning, stacklevel=3)
    with _hide_plugin_deprecation_warnings():
        return call_plugin(
            'imread_collection',
            load_pattern,
            conserve_memory,
            plugin=plugin,
            **plugin_args,
        )


@deprecate_parameter(
    "plugin",
    start_version="0.25",
    stop_version="0.27",
    template=_remove_plugin_param_template,
)
def imsave(fname, arr, plugin=DEPRECATED, *, check_contrast=True, **plugin_args):
    """Save an image to file.

    Parameters
    ----------
    fname : str or pathlib.Path
        Target filename.
    arr : ndarray of shape (M,N) or (M,N,3) or (M,N,4)
        Image data.
    check_contrast : bool, optional
        Check for low contrast and print warning (default: True).

    Other Parameters
    ----------------
    plugin_args : DEPRECATED
        The plugin infrastructure is deprecated.
    """
    if plugin is DEPRECATED:
        plugin = None
    if plugin_args:
        msg = (
            "The plugin infrastructure in `skimage.io` is deprecated since "
            "version 0.25 and will be removed in 0.27 (or later). To avoid "
            "this warning, please do not pass additional keyword arguments "
            "for plugins (`**plugin_args`). Instead, use `imageio` or other "
            "I/O packages directly. See also `skimage.io.imsave`."
        )
        warnings.warn(msg, category=FutureWarning, stacklevel=3)

    if isinstance(fname, pathlib.Path):
        fname = str(fname.resolve())
    if plugin is None and hasattr(fname, 'lower'):
        if fname.lower().endswith(('.tiff', '.tif')):
            plugin = 'tifffile'
    if arr.dtype == bool:
        warn(
            f'{fname} is a boolean image: setting True to 255 and False to 0. '
            'To silence this warning, please convert the image using '
            'img_as_ubyte.',
            stacklevel=3,
        )
        arr = arr.astype('uint8') * 255
    if check_contrast and is_low_contrast(arr):
        warn(f'{fname} is a low contrast image')

    with _hide_plugin_deprecation_warnings():
        return call_plugin('imsave', fname, arr, plugin=plugin, **plugin_args)


@deprecate_func(
    deprecated_version="0.25",
    removed_version="0.27",
    hint="Please use `matplotlib`, `napari`, etc. to visualize images.",
)
def imshow(arr, plugin=None, **plugin_args):
    """Display an image.

    Parameters
    ----------
    arr : ndarray or str
        Image data or name of image file.
    plugin : str
        Name of plugin to use.  By default, the different plugins are
        tried (starting with imageio) until a suitable candidate is found.

    Other Parameters
    ----------------
    plugin_args : keywords
        Passed to the given plugin.

    """
    if isinstance(arr, str):
        arr = call_plugin('imread', arr, plugin=plugin)
    with _hide_plugin_deprecation_warnings():
        return call_plugin('imshow', arr, plugin=plugin, **plugin_args)


@deprecate_func(
    deprecated_version="0.25",
    removed_version="0.27",
    hint="Please use `matplotlib`, `napari`, etc. to visualize images.",
)
def imshow_collection(ic, plugin=None, **plugin_args):
    """Display a collection of images.

    Parameters
    ----------
    ic : :class:`ImageCollection`
        Collection to display.

    Other Parameters
    ----------------
    plugin_args : keywords
        Passed to the given plugin.

    """
    with _hide_plugin_deprecation_warnings():
        return call_plugin('imshow_collection', ic, plugin=plugin, **plugin_args)


@require("matplotlib", ">=3.3")
@deprecate_func(
    deprecated_version="0.25",
    removed_version="0.27",
    hint="Please use `matplotlib`, `napari`, etc. to visualize images.",
)
def show():
    """Display pending images.

    Launch the event loop of the current GUI plugin, and display all
    pending images, queued via `imshow`. This is required when using
    `imshow` from non-interactive scripts.

    A call to `show` will block execution of code until all windows
    have been closed.

    Examples
    --------
    .. testsetup::
        >>> import pytest; _ = pytest.importorskip('matplotlib')

    >>> import skimage.io as io
    >>> rng = np.random.default_rng()
    >>> for i in range(4):
    ...     ax_im = io.imshow(rng.random((50, 50)))  # doctest: +SKIP
    >>> io.show() # doctest: +SKIP

    """
    with _hide_plugin_deprecation_warnings():
        return call_plugin('_app_show')



"""Test for correctness of color distance functions"""

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_almost_equal, assert_equal

from skimage._shared.testing import fetch
from skimage._shared.utils import _supported_float_type
from skimage.color.delta_e import (
    deltaE_cie76,
    deltaE_ciede94,
    deltaE_ciede2000,
    deltaE_cmc,
)


@pytest.mark.parametrize("channel_axis", [0, 1, -1])
@pytest.mark.parametrize('dtype', [np.float32, np.float64])
def test_ciede2000_dE(dtype, channel_axis):
    data = load_ciede2000_data()
    N = len(data)
    lab1 = np.zeros((N, 3), dtype=dtype)
    lab1[:, 0] = data['L1']
    lab1[:, 1] = data['a1']
    lab1[:, 2] = data['b1']

    lab2 = np.zeros((N, 3), dtype=dtype)
    lab2[:, 0] = data['L2']
    lab2[:, 1] = data['a2']
    lab2[:, 2] = data['b2']

    lab1 = np.moveaxis(lab1, source=-1, destination=channel_axis)
    lab2 = np.moveaxis(lab2, source=-1, destination=channel_axis)
    dE2 = deltaE_ciede2000(lab1, lab2, channel_axis=channel_axis)
    assert dE2.dtype == _supported_float_type(dtype)

    assert_allclose(dE2, data['dE'], rtol=1e-2)


def load_ciede2000_data():
    dtype = [
        ('pair', int),
        ('1', int),
        ('L1', float),
        ('a1', float),
        ('b1', float),
        ('a1_prime', float),
        ('C1_prime', float),
        ('h1_prime', float),
        ('hbar_prime', float),
        ('G', float),
        ('T', float),
        ('SL', float),
        ('SC', float),
        ('SH', float),
        ('RT', float),
        ('dE', float),
        ('2', int),
        ('L2', float),
        ('a2', float),
        ('b2', float),
        ('a2_prime', float),
        ('C2_prime', float),
        ('h2_prime', float),
    ]

    # note: ciede_test_data.txt contains several intermediate quantities
    path = fetch('color/tests/ciede2000_test_data.txt')
    return np.loadtxt(path, dtype=dtype)


@pytest.mark.parametrize("channel_axis", [0, 1, -1])
@pytest.mark.parametrize('dtype', [np.float32, np.float64])
def test_cie76(dtype, channel_axis):
    data = load_ciede2000_data()
    N = len(data)
    lab1 = np.zeros((N, 3), dtype=dtype)
    lab1[:, 0] = data['L1']
    lab1[:, 1] = data['a1']
    lab1[:, 2] = data['b1']

    lab2 = np.zeros((N, 3), dtype=dtype)
    lab2[:, 0] = data['L2']
    lab2[:, 1] = data['a2']
    lab2[:, 2] = data['b2']

    lab1 = np.moveaxis(lab1, source=-1, destination=channel_axis)
    lab2 = np.moveaxis(lab2, source=-1, destination=channel_axis)
    dE2 = deltaE_cie76(lab1, lab2, channel_axis=channel_axis)
    assert dE2.dtype == _supported_float_type(dtype)
    oracle = np.array(
        [
            4.00106328,
            6.31415011,
            9.1776999,
            2.06270077,
            2.36957073,
            2.91529271,
            2.23606798,
            2.23606798,
            4.98000036,
            4.9800004,
            4.98000044,
            4.98000049,
            4.98000036,
            4.9800004,
            4.98000044,
            3.53553391,
            36.86800781,
            31.91002977,
            30.25309901,
            27.40894015,
            0.89242934,
            0.7972,
            0.8583065,
            0.82982507,
            3.1819238,
            2.21334297,
            1.53890382,
            4.60630929,
            6.58467989,
            3.88641412,
            1.50514845,
            2.3237848,
            0.94413208,
            1.31910843,
        ]
    )
    rtol = 1e-5 if dtype == np.float32 else 1e-8
    assert_allclose(dE2, oracle, rtol=rtol)


@pytest.mark.parametrize("channel_axis", [0, 1, -1])
@pytest.mark.parametrize('dtype', [np.float32, np.float64])
def test_ciede94(dtype, channel_axis):
    data = load_ciede2000_data()
    N = len(data)
    lab1 = np.zeros((N, 3), dtype=dtype)
    lab1[:, 0] = data['L1']
    lab1[:, 1] = data['a1']
    lab1[:, 2] = data['b1']

    lab2 = np.zeros((N, 3), dtype=dtype)
    lab2[:, 0] = data['L2']
    lab2[:, 1] = data['a2']
    lab2[:, 2] = data['b2']

    lab1 = np.moveaxis(lab1, source=-1, destination=channel_axis)
    lab2 = np.moveaxis(lab2, source=-1, destination=channel_axis)
    dE2 = deltaE_ciede94(lab1, lab2, channel_axis=channel_axis)
    assert dE2.dtype == _supported_float_type(dtype)
    oracle = np.array(
        [
            1.39503887,
            1.93410055,
            2.45433566,
            0.68449187,
            0.6695627,
            0.69194527,
            2.23606798,
            2.03163832,
            4.80069441,
            4.80069445,
            4.80069449,
            4.80069453,
            4.80069441,
            4.80069445,
            4.80069449,
            3.40774352,
            34.6891632,
            29.44137328,
            27.91408781,
            24.93766082,
            0.82213163,
            0.71658427,
            0.8048753,
            0.75284394,
            1.39099471,
            1.24808929,
            1.29795787,
            1.82045088,
            2.55613309,
            1.42491303,
            1.41945261,
            2.3225685,
            0.93853308,
            1.30654464,
        ]
    )
    rtol = 1e-5 if dtype == np.float32 else 1e-8
    assert_allclose(dE2, oracle, rtol=rtol)


@pytest.mark.parametrize("channel_axis", [0, 1, -1])
@pytest.mark.parametrize('dtype', [np.float32, np.float64])
def test_cmc(dtype, channel_axis):
    data = load_ciede2000_data()
    N = len(data)
    lab1 = np.zeros((N, 3), dtype=dtype)
    lab1[:, 0] = data['L1']
    lab1[:, 1] = data['a1']
    lab1[:, 2] = data['b1']

    lab2 = np.zeros((N, 3), dtype=dtype)
    lab2[:, 0] = data['L2']
    lab2[:, 1] = data['a2']
    lab2[:, 2] = data['b2']

    lab1 = np.moveaxis(lab1, source=-1, destination=channel_axis)
    lab2 = np.moveaxis(lab2, source=-1, destination=channel_axis)
    dE2 = deltaE_cmc(lab1, lab2, channel_axis=channel_axis)
    assert dE2.dtype == _supported_float_type(dtype)
    oracle = np.array(
        [
            1.73873611,
            2.49660844,
            3.30494501,
            0.85735576,
            0.88332927,
            0.97822692,
            3.50480874,
            2.87930032,
            6.5783807,
            6.57838075,
            6.5783808,
            6.57838086,
            6.67492321,
            6.67492326,
            6.67492331,
            4.66852997,
            42.10875485,
            39.45889064,
            38.36005919,
            33.93663807,
            1.14400168,
            1.00600419,
            1.11302547,
            1.05335328,
            1.42822951,
            1.2548143,
            1.76838061,
            2.02583367,
            3.08695508,
            1.74893533,
            1.90095165,
            1.70258148,
            1.80317207,
            2.44934417,
        ]
    )
    rtol = 1e-5 if dtype == np.float32 else 1e-8
    assert_allclose(dE2, oracle, rtol=rtol)

    # Equal or close colors make `delta_e.get_dH2` function to return
    # negative values resulting in NaNs when passed to sqrt (see #1908
    # issue on Github):
    lab1 = lab2
    expected = np.zeros_like(oracle)
    assert_almost_equal(
        deltaE_cmc(lab1, lab2, channel_axis=channel_axis), expected, decimal=6
    )

    lab2[0, 0] += np.finfo(float).eps
    assert_almost_equal(
        deltaE_cmc(lab1, lab2, channel_axis=channel_axis), expected, decimal=6
    )


def test_cmc_single_item():
    # Single item case:
    lab1 = lab2 = np.array([0.0, 1.59607713, 0.87755709])
    assert_equal(deltaE_cmc(lab1, lab2), 0)

    lab2[0] += np.finfo(float).eps
    assert_equal(deltaE_cmc(lab1, lab2), 0)


def test_single_color_cie76():
    lab1 = (0.5, 0.5, 0.5)
    lab2 = (0.4, 0.4, 0.4)
    deltaE_cie76(lab1, lab2)


def test_single_color_ciede94():
    lab1 = (0.5, 0.5, 0.5)
    lab2 = (0.4, 0.4, 0.4)
    deltaE_ciede94(lab1, lab2)


def test_single_color_ciede2000():
    lab1 = (0.5, 0.5, 0.5)
    lab2 = (0.4, 0.4, 0.4)
    deltaE_ciede2000(lab1, lab2)


def test_single_color_cmc():
    lab1 = (0.5, 0.5, 0.5)
    lab2 = (0.4, 0.4, 0.4)
    deltaE_cmc(lab1, lab2)
